"""Contains classes and methods to store the solution of an NLP problem after a call
to :meth:`csnlp.Nlp.solve` or :meth:`csnlp.multistart.MultistartNlp.solve_multi`."""

from collections.abc import Iterable as _Iterable
from functools import cached_property as _cached_property
from itertools import product as _product
from typing import TYPE_CHECKING
from typing import Any as _Any
from typing import Optional
from typing import Protocol as _Protocol
from typing import TypeVar as _TypeVar
from typing import Union

import casadi as cs
import numpy as np
import numpy.typing as npt

from .data import array2cs as _array2cs
from .data import cs2array as _cs2array

if TYPE_CHECKING:
    from ..nlps.nlp import Nlp

SymType = _TypeVar("SymType", cs.SX, cs.MX)
SymOrNumType = _TypeVar("SymOrNumType", cs.SX, cs.MX, cs.DM, int, float, np.ndarray)


def _is_infeas(status: str, solver_plugin: str) -> Optional[bool]:
    """Internal utility to compute whether the solver status indicates infeasibility."""
    # NLPs
    if solver_plugin == "ipopt":
        return status == "Infeasible_Problem_Detected"
    if solver_plugin in ("qrsqp", "sqpmethod"):
        return status == "Search_Direction_Becomes_Too_Small"
    # QPs
    if solver_plugin == "osqp":
        return "infeasible" in status
    if solver_plugin == "proxqp":
        return (
            status == "PROXQP_PRIMAL_INFEASIBLE" or status == "PROXQP_DUAL_INFEASIBLE"
        )
    if solver_plugin == "qpoases":
        return "infeasib" in status
    if solver_plugin == "qrqp":
        return status == "Failed to calculate search direction"
    # LPs
    if solver_plugin == "clp":
        return status.endswith("infeasible")
    # MIPs
    if solver_plugin in ("bonmin", "gurobi"):
        return status == "INFEASIBLE"
    if solver_plugin == "cbc":
        return "not feasible" in status
    if solver_plugin == "gurobi":
        return status == "INFEASIBLE" or status == "INF_OR_UNBD"
    if solver_plugin == "knitro":
        return "INFEAS" in status
    return None


class Solution(_Protocol[SymType]):
    """Class containing information on the solution of a solver's run for an instance of
    :class:`csnlp.Nlp`.

    Notes
    -----
    This class is merely a protocol, and as such it just defines the interface of how a
    solution should look like, plus some minor implementations.
    """

    @property
    def f(self) -> float:
        """Optimal value of the objective function."""

    @property
    def p_sym(self) -> cs.DM:
        """Symbolical  parameters in a vector."""

    @property
    def p(self) -> cs.DM:
        """Values of the parameters for which this solver's solution was generated."""

    @property
    def x_sym(self) -> SymType:
        """Symbolical primal variables in a vector."""

    @property
    def x(self) -> cs.DM:
        """Optimal values of the primal variables in a vector."""

    @property
    def lam_g_and_h_sym(self) -> cs.DM:
        """Symbolical the equality and inequality dual variables in a vector."""

    @property
    def lam_g_and_h(self) -> cs.DM:
        """Optimal values of the equality and inequality dual variables in a vector."""

    @property
    def lam_lbx_and_ubx_sym(self) -> cs.DM:
        """Symbolical lower- and upper-bounds dual variables in a vector."""

    @property
    def lam_lbx_and_ubx(self) -> cs.DM:
        """Optimal values of the lower- and upper-bounds dual variables in a vector."""

    @_cached_property
    def x_and_lam_and_p_sym(self) -> SymType:
        """Symbolical primal and dual variables and parameters in a vector."""
        return cs.vertcat(
            self.x_sym,
            self.lam_g_and_h_sym,
            self.lam_lbx_and_ubx_sym,
            self.p_sym,
        )

    @_cached_property
    def x_and_lam_and_p(self) -> cs.DM:
        """Optimal values of the primal and dual variables, as well as the parameters
        for which the solution was found, in a vector."""
        return cs.vertcat(
            self.x,
            self.lam_g_and_h,
            self.lam_lbx_and_ubx,
            self.p,
        )

    @property
    def vars(self) -> dict[str, SymType]:
        """Symbolical primal variables."""
        return self._vars

    @property
    def vals(self) -> dict[str, cs.DM]:
        """Optimal values of the primal variables."""

    @property
    def dual_vars(self) -> dict[str, SymType]:
        """Symbolical dual variables."""
        return self._dual_vars

    @property
    def dual_vals(self) -> dict[str, cs.DM]:
        """Optimal values of the dual variables."""

    @property
    def solver_plugin(self) -> str:
        """The solver plugin used to generate this solution."""
        return self._solver_plugin

    @property
    def stats(self) -> dict[str, _Any]:
        """Statistics of the solver for this solution's run."""
        return self._stats

    @property
    def status(self) -> str:
        """Gets the status of the solver at this solution."""
        return self.stats["return_status"]

    @property
    def unified_return_status(self) -> str:
        """Gets the unified status of the solver at this solution."""
        return self.stats["unified_return_status"]

    @property
    def success(self) -> bool:
        """Gets whether the solver's run was successful."""
        return self.stats["success"]

    @_cached_property
    def infeasible(self) -> Optional[bool]:
        r"""Gets whether the solver status indicates infeasibility. If ``False``, it
        does not imply feasibility as the solver or its CasADi interface may have not
        detect it.

        For different solvers, the infeasibility status is stored in different ways.
        Here is a list of what I gathered so far. The solvers are grouped based on the
        type of problem they solve. An (F) next to the solver's name indicates that the
        the solver will crash the program if ``"error_on_fail": True`` and the solver
        fails. The ``status`` and ``unified_return_status`` can both be found in
        the solver's stats, or in this solution object.

        * NLPs

          - **fatrop**: unclear; better to return ``None`` for now
          - **ipopt**: ``status == "Infeasible_Problem_Detected"``
          - **qrsqp** (F): ``status == "Search_Direction_Becomes_Too_Small"``
            (dubious)
          - **sqpmethod** (F): ``status == "Search_Direction_Becomes_Too_Small"``
            (dubious)

        * QPs

          - **ipqp** (F): no clear way to detect infeasibility; return ``None`` for now
          - **osqp** (F): ``unified_return_status == "SOLVER_RET_INFEASIBLE"`` or
            ``"infeasible" in status``
          - **proxqp** (F): ``status == "PROXQP_PRIMAL_INFEASIBLE"`` or
            ``status == "PROXQP_DUAL_INFEASIBLE"``
          - **qpoases** (F): ``"infeasib" in status``
          - **qrqp** (F): ``status == "Failed to calculate search direction"``

        * LPs

          - **clp** (F): ``status == "primal infeasible"`` or
            ``status == "dual infeasible"``

        * Mixed-Iteger Problems (MIPs)

          - **bonmin** (F): ``status == "INFEASIBLE"``
          - **cbc** (F): ``"not feasible" in status``
          - **gurobi** (F): ``status == "INFEASIBLE"`` or ``status == "INF_OR_UNBD"``
          - **knitro**: ``"INFEAS" in status``
        """
        return _is_infeas(self.status, self.solver_plugin)

    @property
    def barrier_parameter(self) -> float:
        """Gets the IPOPT barrier parameter at the optimal solution"""
        return self.stats["iterations"]["mu"][-1]

    def value(
        self, expr: Union[SymType, np.ndarray], eval: bool = True
    ) -> Union[SymType, cs.DM]:
        """Computes the value of the expression substituting the values of this
        solution in the expression.

        Parameters
        ----------
        expr : casadi.SX, MX or an array of these
            The symbolic expression to be evaluated at the solution's values.
        eval : bool, optional
            Evaluates numerically the new expression. By default, ``True``. See
            :meth:`csnlp.solutions.subsevalf` for more details.

        Returns
        -------
        casadi.SX or MX or DM
            The expression evaluated with the solution's values.

        Raises
        ------
        RuntimeError
            Raises if ``eval=True`` but there are symbolic variables that are still
            free. This can occur when there are symbols that are outside the solution's
            variables, and thus have not been substituted by a numerical value.
        """
        return subsevalf(
            expr, self.x_and_lam_and_p_sym, self.x_and_lam_and_p, eval=eval
        )

    @staticmethod
    def from_casadi_solution(
        sol_with_stats: dict[str, _Any], nlp: "Nlp[SymType]"
    ) -> "Solution[SymType]":
        """Creates a new solution from a CasADi solution,

        Parameters
        ----------
        sol_with_stats : dict of (str, cs.DM) and one entry with stats, i.e., Any
            The solution dictionary from the CasADi solver, which contains the optimal
            values of the primal and dual variables, as well as the parameters, and the
            solver's statistics.
        nlp : Nlp[SymType]
            The NLP instance for which the solution was computed.

        Returns
        -------
        Solution[SymType]
            The solution corresponding to the CasADi solution.
        """

    @staticmethod
    def cmp_key(sol: "Solution[SymType]") -> tuple[bool, bool, float]:
        """Gets the comparison keys to compare a solution with another. This solution is
        strictly better if

        - it is feasible and the other is not, or
        - both are feasible or infeasible, and the current is successful and the other
            is not, or
        - both are successful or not, and the current has a lower optimal value than the
            other.

        To be used as ``key`` argument in, e.g., :func:`min` or :func:`sorted`.

        Returns
        -------
        tuple of (bool, bool, float)
            A tuple with (is_infeasible, is_unsuccessful, f).
        """
        is_infeas = _is_infeas(sol.status, sol.solver_plugin)
        if is_infeas is None:
            is_infeas = "infeas" in sol.status.lower()
        return is_infeas, not sol.success, sol.f

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(f={self.f},success={self.success},"
            f"status={self.status})"
        )


class EagerSolution(Solution[SymType]):
    """Class containing information on the solution of a solver's run for an instance of
    :class:`csnlp.Nlp`.

    Parameters
    ----------
    f : float
        Optimal value of the NLP at the solution.
    p_sym : cs.SX or MX
        Symbolic parameters in a vector.
    p : cs.DM
        Values of the parameters for which this solver's solution was generated.
    x_sym : SymType
        Symbolic primal variables in a vector.
    x : cs.DM
        Optimal values of the primal variables in a vector.
    lam_g_and_h_sym : cs.SX or MX
        Symbolic the equality and inequality dual variables in a vector.
    lam_g_and_h : cs.DM
        Optimal values of the equality and inequality dual variables in a vector.
    lam_lbx_and_ubx_sym : cs.SX or MX
        Symbolic lower- and upper-bounds dual variables in a vector.
    lam_lbx_and_ubx : cs.DM
        Optimal values of the lower- and upper-bounds dual variables in a vector.
    vars : dict of (str, cs.SX or MX)
        Symbolic primal variables.
    vals : dict of (str, cs.DM)
        Optimal values of the primal variables.
    dual_vars : dict of (str, cs.SX or MX)
        Symbolic dual variables.
    dual_vals : dict of (str, cs.DM)
        Optimal values of the dual variables.
    stats : dict
        Stats of the solver run that generated this solution.

    Notes
    -----
    This class is an eager implementation of the :class:`csnlp.solutions.Solution`,
    where the values and properties of the solution are computed eagerly after the
    solver has finished its run.
    """

    def __init__(
        self,
        f: float,
        p_sym: SymType,
        p: cs.DM,
        x_sym: SymType,
        x: cs.DM,
        lam_g_and_h_sym: SymType,
        lam_g_and_h: cs.DM,
        lam_lbx_and_ubx_sym: SymType,
        lam_lbx_and_ubx: cs.DM,
        vars: dict[str, SymType],
        vals: dict[str, cs.DM],
        dual_vars: dict[str, SymType],
        dual_vals: dict[str, cs.DM],
        stats: dict[str, _Any],
        solver_plugin: str,
    ) -> None:
        self._f = f

        self._p_sym = p_sym
        self._p = p
        self._x_sym = x_sym
        self._x = x
        self._lam_g_and_h_sym = lam_g_and_h_sym
        self._lam_g_and_h = lam_g_and_h
        self._lam_lbx_and_ubx_sym = lam_lbx_and_ubx_sym
        self._lam_lbx_and_ubx = lam_lbx_and_ubx

        self._vars = vars
        self._vals = vals
        self._dual_vars = dual_vars
        self._dual_vals = dual_vals

        self._stats = stats
        self._solver_plugin = solver_plugin

    @property
    def f(self) -> float:
        return self._f

    @property
    def p_sym(self) -> cs.DM:
        return self._p_sym

    @property
    def p(self) -> cs.DM:
        return self._p

    @property
    def x_sym(self) -> SymType:
        return self._x_sym

    @property
    def x(self) -> cs.DM:
        return self._x

    @property
    def lam_g_and_h_sym(self) -> cs.DM:
        return self._lam_g_and_h_sym

    @property
    def lam_g_and_h(self) -> cs.DM:
        return self._lam_g_and_h

    @property
    def lam_lbx_and_ubx_sym(self) -> cs.DM:
        return self._lam_lbx_and_ubx_sym

    @property
    def lam_lbx_and_ubx(self) -> cs.DM:
        return self._lam_lbx_and_ubx

    @property
    def vals(self) -> dict[str, cs.DM]:
        return self._vals

    @property
    def dual_vals(self) -> dict[str, cs.DM]:
        return self._dual_vals

    @staticmethod
    def from_casadi_solution(
        sol_with_stats: dict[str, _Any], nlp: "Nlp[SymType]"
    ) -> "EagerSolution[SymType]":
        # stats and objective
        stats = sol_with_stats.pop("stats")
        sol: dict[str, cs.DM] = sol_with_stats  # now the solution has only cs.DMs
        f = float(sol["f"])

        # primal variables and values
        x = nlp._x
        x_opt = sol["x"]
        vars = nlp._vars.copy()
        vals = {name: subsevalf(var, x, x_opt) for name, var in vars.items()}

        # dual variables and values
        lam_g_and_h = sol["lam_g"]
        lam_g = lam_g_and_h[: nlp.ng, :]
        lam_h = lam_g_and_h[nlp.ng :, :]
        lam_lbx = -cs.fmin(sol["lam_x"][nlp.nonmasked_lbx_idx, :], 0)
        lam_ubx = cs.fmax(sol["lam_x"][nlp.nonmasked_ubx_idx, :], 0)
        dual_vars = nlp._dual_vars.copy()
        dual_vals = {}
        for n, var in dual_vars.items():
            if n.startswith("lam_g"):
                dual_vals[n] = subsevalf(var, nlp._lam_g, lam_g)
            elif n.startswith("lam_h"):
                dual_vals[n] = subsevalf(var, nlp._lam_h, lam_h)
            elif n.startswith("lam_lb"):
                dual_vals[n] = subsevalf(var, nlp._lam_lbx, lam_lbx)
            elif n.startswith("lam_ub"):
                dual_vals[n] = subsevalf(var, nlp._lam_ubx, lam_ubx)
            else:
                raise RuntimeError(f"unknown dual variable type {n}")

        return EagerSolution(
            f,
            nlp._p,
            sol["p"],
            x,
            x_opt,
            cs.vertcat(nlp._lam_g, nlp._lam_h),
            lam_g_and_h,
            cs.vertcat(nlp._lam_lbx, nlp._lam_ubx),
            cs.vertcat(lam_lbx, lam_ubx),
            vars,
            vals,
            dual_vars,
            dual_vals,
            stats,
            nlp.unwrapped._solver_plugin,
        )


class LazySolution(Solution[SymType]):
    """Class containing information on the solution of a solver's run for an instance of
    :class:`csnlp.Nlp`.

    Parameters
    ----------
    f : float
        Optimal value of the NLP at the solution.
    p_sym : cs.SX or MX
        Symbolic parameters in a vector.
    x_sym : SymType
        Symbolic primal variables in a vector.
    lam_g_and_h_sym : cs.SX or MX
        Symbolic the equality and inequality dual variables in a vector.
    lam_lbx_and_ubx_sym : cs.SX or MX
        Symbolic lower- and upper-bounds dual variables in a vector.
    vars : dict of (str, cs.SX or MX)
        Symbolic primal variables.
    dual_vars : dict of (str, cs.SX or MX)
        Symbolic dual variables.
    nonmasked_lbx_idx : slice or 1D array of int
        Indexes of the non-masked (i.e., valid, finite) lower-bounds.
    nonmasked_ubx_idx : slice or 1D array of int
        Indexes of the non-masked (i.e., valid, finite) upper-bounds.
    stats : dict
        Stats of the solver run that generated this solution.

    Notes
    -----
    This class is a lazy implementation of the :class:`csnlp.solutions.Solution`, where
    the values and properties of the solution are computed lazily only when requested.
    """

    def __init__(
        self,
        solution: dict[str, cs.DM],
        p_sym: SymType,
        x_sym: SymType,
        lam_g_and_h_sym: SymType,
        lam_lbx_and_ubx_sym: SymType,
        vars: dict[str, SymType],
        dual_vars: dict[str, SymType],
        nonmasked_lbx_idx: Union[slice, npt.NDArray[np.int64]],
        nonmasked_ubx_idx: Union[slice, npt.NDArray[np.int64]],
        stats: dict[str, _Any],
        solver_plugin: str,
    ) -> None:
        self._sol = solution
        self._p_sym = p_sym
        self._x_sym = x_sym
        self._lam_g_and_h_sym = lam_g_and_h_sym
        self._lam_lbx_and_ubx_sym = lam_lbx_and_ubx_sym
        self._vars = vars
        self._dual_vars = dual_vars
        self._nonmasked_lbx_idx = nonmasked_lbx_idx
        self._nonmasked_ubx_idx = nonmasked_ubx_idx
        self._stats = stats
        self._solver_plugin = solver_plugin

    @_cached_property
    def f(self) -> float:
        return float(self._sol["f"])

    @property
    def p_sym(self) -> SymType:
        return self._p_sym

    @_cached_property
    def p(self) -> cs.DM:
        return self._sol["p"]

    @property
    def x_sym(self) -> SymType:
        return self._x_sym

    @_cached_property
    def x(self) -> cs.DM:
        return self._sol["x"]

    @property
    def lam_g_and_h_sym(self) -> cs.DM:
        return self._lam_g_and_h_sym

    @_cached_property
    def lam_g_and_h(self) -> cs.DM:
        return self._sol["lam_g"]

    @property
    def lam_lbx_and_ubx_sym(self) -> cs.DM:
        return self._lam_lbx_and_ubx_sym

    @_cached_property
    def lam_lbx_and_ubx(self) -> cs.DM:
        lam_lbx = -cs.fmin(self._sol["lam_x"][self._nonmasked_lbx_idx, :], 0)
        lam_ubx = cs.fmax(self._sol["lam_x"][self._nonmasked_ubx_idx, :], 0)
        return cs.vertcat(lam_lbx, lam_ubx)

    @_cached_property
    def vals(self) -> dict[str, cs.DM]:
        x = self._x_sym
        x_vals = self._sol["x"]
        return {n: cs.evalf(cs.substitute(v, x, x_vals)) for n, v in self._vars.items()}

    @_cached_property
    def dual_vals(self) -> dict[str, cs.DM]:
        lam_g_and_h_sym = self.lam_g_and_h_sym
        lam_g_and_h = self.lam_g_and_h
        lam_lbx_and_ubx_sym = self.lam_lbx_and_ubx_sym
        lam_lbx_and_ubx = self.lam_lbx_and_ubx
        dual_vals = {}
        for n, v in self._dual_vars.items():
            if n.startswith("lam_g") or n.startswith("lam_h"):
                dual_vals[n] = cs.evalf(cs.substitute(v, lam_g_and_h_sym, lam_g_and_h))
            elif n.startswith("lam_lb") or n.startswith("lam_ub"):
                dual_vals[n] = cs.evalf(
                    cs.substitute(v, lam_lbx_and_ubx_sym, lam_lbx_and_ubx)
                )
            else:
                raise RuntimeError(f"unknown dual variable type `{n}`")
        return dual_vals

    @staticmethod
    def from_casadi_solution(
        sol_with_stats: dict[str, _Any], nlp: "Nlp[SymType]"
    ) -> "LazySolution[SymType]":
        stats = sol_with_stats.pop("stats")
        sol: dict[str, cs.DM] = sol_with_stats
        x = nlp._x
        p = nlp._p
        lam_g_and_h = cs.vertcat(nlp._lam_g, nlp._lam_h)
        lam_lbx_and_ubx = cs.vertcat(nlp._lam_lbx, nlp._lam_ubx)
        vars = nlp._vars.copy()
        dual_vars = nlp._dual_vars.copy()
        idx = nlp.nonmasked_lbx_idx
        nonmasked_lbx_idx = idx.copy() if isinstance(idx, np.ndarray) else idx
        idx = nlp.nonmasked_ubx_idx
        nonmasked_ubx_idx = idx.copy() if isinstance(idx, np.ndarray) else idx
        return LazySolution(
            sol,
            p,
            x,
            lam_g_and_h,
            lam_lbx_and_ubx,
            vars,
            dual_vars,
            nonmasked_lbx_idx,
            nonmasked_ubx_idx,
            stats,
            nlp.unwrapped._solver_plugin,
        )


def _broadcast_like(x: SymOrNumType, other: SymOrNumType) -> Union[SymType, np.ndarray]:
    """Internal utility to broadcast a value, if numerical, to the other's shape."""
    if isinstance(x, (np.ndarray, cs.DM)):
        target_shape = other.shape
        if x.shape == target_shape or (
            other.is_vector()
            and (x.size if isinstance(x, np.ndarray) else x.numel()) == other.numel()
        ):
            return x
        return np.broadcast_to(x, target_shape)
    if isinstance(x, (int, float)):
        return np.full(other.shape, x)
    return x  # cs.SX or cs.MX


def _internal_subsevalf_cs(
    expr: SymType,
    old: Union[SymType, dict[str, SymType], _Iterable[SymType]],
    new: Union[SymOrNumType, dict[str, SymOrNumType], _Iterable[SymOrNumType]],
    eval: bool,
) -> Union[SymType, cs.DM]:
    """Internal utility for substituting and evaluting casadi objects."""
    if isinstance(expr, cs.DM):
        return expr
    if isinstance(old, (cs.SX, cs.MX)):
        new_expr = cs.substitute(expr, old, new)
    else:
        old_vals = []
        new_vals = []
        if isinstance(old, dict):
            assert isinstance(new, dict), "old and new should be of the same type"
            for name, old_val in old.items():
                new_val = new[name]
                old_vals.append(old_val)
                new_vals.append(_broadcast_like(new_val, old_val))
        else:  # iterable
            for old_val, new_val in zip(old, new):
                old_vals.append(old_val)
                new_vals.append(_broadcast_like(new_val, old_val))
        new_expr = cs.substitute(expr, cs.vvcat(old_vals), cs.vvcat(new_vals))
    return cs.evalf(new_expr) if eval else new_expr


def _internal_subsevalf_np(
    expr: np.ndarray,
    old: Union[SymType, dict[str, SymType], _Iterable[SymType]],
    new: Union[SymOrNumType, dict[str, SymOrNumType], _Iterable[SymOrNumType]],
    eval: bool,
) -> Union[SymType, np.ndarray, cs.DM]:
    """Internal utility for substituting and evaluting arrays of casadi objects."""
    # if not symbolic, return it
    if expr.dtype != object:
        return expr

    # up to 2D, we can get away with only one substitution
    if expr.ndim <= 2:
        return _internal_subsevalf_cs(_array2cs(expr), old, new, eval)

    # for tensors, check if the right end or the left is bigger, then loop over the
    # rest while substituing in the right/left end of the array
    shape = expr.shape
    transposed = (shape[0] * shape[1]) > (shape[-2] * shape[-1])
    if transposed:
        shape = shape[::-1]
        expr = expr.transpose()
    out = np.empty(shape, object)
    shape_iter, shape_cs = shape[:-2], shape[-2:]
    for i in _product(*map(range, shape_iter)):
        out[i] = _cs2array(
            _internal_subsevalf_cs(_array2cs(expr[i]), old, new, eval)
        ).reshape(shape_cs)
    if transposed:
        out = out.transpose()
        expr = expr.transpose()

    if eval:
        out = out.astype(float)
    return out


def subsevalf(
    expr: Union[SymType, np.ndarray],
    old: Union[SymType, dict[str, SymType], _Iterable[SymType]],
    new: Union[SymOrNumType, dict[str, SymOrNumType], _Iterable[SymOrNumType]],
    eval: bool = True,
) -> Union[SymType, cs.DM, np.ndarray]:
    """Substitutes the old variables with the new ones in the symbolic expression,
    and evaluates it, if required.

    Parameters
    ----------
    expr : casadi.SX, MX or an array of these
        Expression for substitution and, possibly, evaluation.
    old : casadi.SX or MX or dict/iterable of these
        Old variable to be substituted.
    new : numpy.array or casadi.SX, MX, DM, int, float, or dict/iterable of these
        New variable that substitutes the old one. If a collection, it is
        assumed the type is the same of ``old`` (so, old and new should share
        collection type), e.g., if ``old`` is a dict, ``new`` should be a dict.
    eval : bool, optional
        Evaluates numerically the new expression. By default, ``True``.

    Returns
    -------
    new_expr : casadi.SX, MX, DM or an array of these
        New expression after substitution (SX, MX) and, possibly, evaluation
        (DM).

    Raises
    ------
    Exception
        Raises if ``old`` and ``new`` are not compatible with :func:`casadi.substitute`;
        or if ``eval=True`` but there are symbolic variables that are still free, i.e.,
        the expression cannot be evaluated numerically since it is still (partially)
        symbolic.
    """
    if isinstance(expr, np.ndarray):
        return _internal_subsevalf_np(expr, old, new, eval)
    else:
        return _internal_subsevalf_cs(expr, old, new, eval)



"""A collection of methods to handle caching in the package. In particular, it offers a
decorator :func:`invalidate_cache` that allows to invalidate the cache of a given set of
other cached properties or methods when the decorated method is invoked, as well as a
function :func:`invalidate_caches_of` that allows to invalidate the cache of a given
object on the fly."""

import functools
import inspect
from typing import Callable


def _is_cached_property(c: Callable) -> bool:
    """Returns True if the callable is a cached property."""
    return isinstance(c, functools.cached_property)


def _is_lru_cache(c: Callable) -> bool:
    """Returns True if the callable is a lru cache."""
    return hasattr(c, "cache_info") or hasattr(c, "cache_clear")


def invalidate_cache(*callables: Callable) -> Callable:
    r"""Decorator that allows to enhance a function or method with the ability, when
    called, to invalidate and clear the cached of some other target methods/properties.
    This is especially useful to reset the cache of a given cached method/property when
    another method makes changes to the underlying data, thus compromising the cached
    results.

    The decorator works only for the built-in :func:`functools.cached_property` and
    :func:`functools.lru_cache`.

    Parameters
    ----------
    callables : cached_property or lru cache wrapper
        The cached properties or methods to be reset in this decorator.

    Returns
    -------
    decorating_function : Callable
        Returns the function wrapped with this decorator.

    Raises
    ------
    ValueError
        Raises if no callable is passed to the function.
    TypeError
        Raises if the given inputs are not instances of
        :func:`functools.cached_property` or :func:`functools.lru_cache`.

    Notes
    -----
    The wrapper can invalidate other :func:`functools.cached_property` instances, but
    for doing so it assumes the instance of the object (which the property to invalidate
    belongs to) is the first argument of the wrapped method. For lru caches the issue
    does not subsist.
    """
    if not callables:
        raise ValueError("No callables were passed for cache invalidation.")
    cached_properties: list[functools.cached_property] = []
    lru_caches: list[functools._lru_cache_wrapper] = []
    for p in callables:
        if _is_cached_property(p):
            cached_properties.append(p)
        elif _is_lru_cache(p):
            lru_caches.append(p)
        else:
            raise TypeError(
                "Expected cached properties or lru wrappers; got "
                f"{p.__class__.__name__} instead."
            )

    Ncp = len(cached_properties)
    if Ncp == 0:
        invalidate_cached_properties = None
    elif Ncp == 1:
        prop = cached_properties[0]

        def invalidate_cached_properties(self):
            propname = prop.attrname
            if propname in self.__dict__:
                del self.__dict__[propname]

    else:

        def invalidate_cached_properties(self):
            for prop in cached_properties:
                propname = prop.attrname
                if propname in self.__dict__:
                    del self.__dict__[propname]

    Nlru = len(lru_caches)
    if Nlru == 0:
        invalidate_lru_caches = None
    elif Nlru == 1:
        lru_cache = lru_caches[0]

        def invalidate_lru_caches():
            lru_cache.cache_clear()

    else:

        def invalidate_lru_caches():
            for lru_cache in lru_caches:
                lru_cache.cache_clear()

    def decorating_function(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if invalidate_cached_properties is not None and args:
                invalidate_cached_properties(args[0])
            if invalidate_lru_caches is not None:
                invalidate_lru_caches()
            return func(*args, **kwargs)

        return wrapper

    return decorating_function


def invalidate_caches_of(obj) -> None:
    """Similar to the decorator :func:`invalidate_cache`, but clears the case of
    the given object only once.

    Parameters
    ----------
    obj : Any
        The object whose caches are to be cleared.
    """
    # basically do again what csnlp.core.cache.invalidate_cache does
    for membername, member in inspect.getmembers(
        type(obj), predicate=lambda m: _is_cached_property(m) or _is_lru_cache(m)
    ):
        if _is_cached_property(member):
            if membername in obj.__dict__:
                del obj.__dict__[membername]
        elif _is_lru_cache(member):
            getattr(obj, membername).cache_clear()



from collections.abc import Iterable, Iterator
from functools import lru_cache
from itertools import repeat
from typing import Any, ClassVar, Generic, Literal, Optional, TypeVar, Union

import casadi as cs
import numpy as np
import numpy.typing as npt
from joblib import Parallel, delayed
from joblib.memory import MemorizedFunc

from ..core.cache import invalidate_cache
from ..core.solutions import (
    EagerSolution,
    LazySolution,
    Solution,
    _is_infeas,
    subsevalf,
)
from ..nlps.nlp import Nlp
from ..nlps.objective import _solve_and_get_stats

SymType = TypeVar("SymType", cs.SX, cs.MX)


def _n(sym_name: str, scenario: int) -> str:
    """Internal utility for the naming convention of the ``i``-th scenario's symbols."""
    return f"{sym_name}__{scenario}"


def _chained_substitute(
    expr: SymType,
    old_vars: dict[str, SymType],
    new_vars: dict[str, SymType],
    old_pars: dict[str, SymType],
    new_pars: dict[str, SymType],
    old_dual_vars: Optional[dict[str, SymType]] = None,
    new_dual_vars: Optional[dict[str, SymType]] = None,
) -> Union[SymType, cs.DM, np.ndarray]:
    """Internal utility to perform substitutions on variables, dual variables and
    parameters in chain."""
    old = []
    new = []
    for n in old_vars:
        old.append(old_vars[n])
        new.append(new_vars[n])
    for n in old_pars:
        old.append(old_pars[n])
        new.append(new_pars[n])
    if old_dual_vars is not None:
        assert new_dual_vars is not None, "Both old and new dual vars must be given."
        for n in old_dual_vars:
            old.append(old_dual_vars[n])
            new.append(new_dual_vars[n])
    return cs.substitute(expr, cs.vvcat(old), cs.vvcat(new))


def _cmp_key(sol: dict[str, Any], plugin_solver: str) -> tuple[bool, bool, float]:
    """Internal utility, similar to :func:`Solution.cmp_key`, but for native CasADi's
    solution dictionaries."""
    stats = sol["stats"]
    status = stats["return_status"]
    is_infeas = _is_infeas(status, plugin_solver)
    if is_infeas is None:
        is_infeas = "infeas" in status.lower()
    return is_infeas, not stats["success"], float(sol["f"])


class MultistartNlp(Nlp[SymType], Generic[SymType]):
    """Base class for NLP with multistarting. This class lays the foundation for solving
    an NLP problem (described as an instance of :class:`csnlp.Nlp`) multiple times with
    different initial conditions, and requires subclasses to implement the actual
    multistarting logic in :meth:`solve_multi`.

    Parameters
    ----------
    args, kwargs
        See inherited :meth:`csnlp.Nlp.__init__`.
    starts : int
        A positive integer for the number of multiple starting guesses to optimize.

    Raises
    ------
    ValueError
        Raises if the scenario number is invalid.
    """

    is_multi: ClassVar[bool] = True
    """Flag to indicate that this is a multistart NLP."""

    def __init__(self, *args: Any, starts: int, **kwargs: Any) -> None:
        if starts <= 0:
            raise ValueError("Number of scenarios must be positive and > 0.")
        super().__init__(*args, **kwargs)
        self._starts = starts

    def __call__(
        self,
        pars: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        **kwargs: Any,
    ) -> Union[Solution[SymType], list[Solution[SymType]]]:
        # call solve_multi only if either pars or vals0 is an iterable; otherwise, run
        # the single, base NLP
        if (pars is None or isinstance(pars, dict)) and (
            vals0 is None or isinstance(vals0, dict)
        ):
            return self.solve(pars, vals0)
        return self.solve_multi(pars, vals0, **kwargs)

    @property
    def starts(self) -> int:
        """Gets the number of starts."""
        return self._starts

    def solve_multi(
        self,
        pars: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        return_all_sols: bool = False,
        **_,
    ) -> Union[Solution[SymType], list[Solution[SymType]]]:
        """Solves the NLP with multiple initial conditions.

        Parameters
        ----------
        pars : dict of (str, array_like) or iterable of, optional
            An iterable that, for each multistart, contains a dictionary with, for each
            parameter in the NLP scheme, the corresponding numerical value. In case a
            single dict is passed, the same is used across all scenarions. Can be
            ``None`` if no parameters are present.
        vals0 : dict of (str, array_like) or iterable of, optional
            An iterable that, for each multistart, contains a dictionary with, for each
            variable in the NLP scheme, the corresponding initial guess. In case a
            single dict is passed, the same is used across all scenarions. By default
            ``None``, in which case  initial guesses are not passed to the solver.
        return_all_sols : bool, optional
            If ``True``, returns the solution of each multistart of the NLP; otherwise,
            only the best solution is returned. By default, ``False``.

        Returns
        -------
        Solution or list of Solutions
            Depending on the flags ``return_all_sols``, returns

            - the best solution out of all multiple starts
            - all the solutions (one per start).
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement `solve_multi`"
        )


class StackedMultistartNlp(MultistartNlp[SymType], Generic[SymType]):
    """A class that models and solves an NLP problem from multiple starting initial
    guesses by automatically stacking the original problem multiple independent times
    in the same, larger-scale NLP. This allows to solve the original problem multiple
    times via a single call to the solver."""

    def __init__(self, *args: Any, starts: int, **kwargs: Any) -> None:
        # this class essentially is a facade that hides an internal nlp in which the
        # problem (variables, parameters, etc.) are duplicated by the requested number
        # of multiple starts. For this reason, all methods are overridden to create
        # multiples of these in the hidden nlp.
        super().__init__(*args, starts=starts, **kwargs)
        self._stacked_nlp = Nlp(*args, **kwargs)  # actual nlp

    @lru_cache(maxsize=32)
    def _vars_i(self, i: int) -> dict[str, SymType]:
        """Internal utility to retrieve the variables of the ``i``-th scenario."""
        nlp = self._stacked_nlp.unwrapped
        return {k: nlp._vars[_n(k, i)] for k in self._vars}

    @lru_cache(maxsize=32)
    def _pars_i(self, i: int) -> dict[str, SymType]:
        """Internal utility to retrieve the parameters of the ``i``-th scenario."""
        nlp = self._stacked_nlp.unwrapped
        return {k: nlp._pars[_n(k, i)] for k in self._pars}

    @lru_cache(maxsize=32)
    def _dual_vars_i(self, i: int) -> dict[str, SymType]:
        """Internal utility to retrieve the dual variables of the ``i``-th scenario."""
        nlp = self._stacked_nlp.unwrapped
        return {k: nlp._dual_vars[_n(k, i)] for k in self._dual_vars}

    @invalidate_cache(_pars_i)
    def parameter(self, name: str, shape: tuple[int, int] = (1, 1)) -> SymType:
        out = super().parameter(name, shape)
        for i in range(self._starts):
            self._stacked_nlp.parameter(_n(name, i), shape)
        return out

    @invalidate_cache(_vars_i, _dual_vars_i)
    def variable(
        self,
        name: str,
        shape: tuple[int, int] = (1, 1),
        discrete: bool = False,
        lb: Union[npt.ArrayLike, cs.DM] = -np.inf,
        ub: Union[npt.ArrayLike, cs.DM] = +np.inf,
    ) -> tuple[SymType, SymType, SymType]:
        out = super().variable(name, shape, discrete, lb, ub)
        for i in range(self._starts):
            self._stacked_nlp.variable(_n(name, i), shape, discrete, lb, ub)
        return out

    @invalidate_cache(_vars_i, _dual_vars_i)
    def constraint(
        self,
        name: str,
        lhs: Union[SymType, np.ndarray, cs.DM],
        op: Literal["==", ">=", "<="],
        rhs: Union[SymType, np.ndarray, cs.DM],
        soft: bool = False,
        simplify: bool = True,
    ) -> tuple[SymType, ...]:
        expr = lhs - rhs
        if simplify:
            expr = cs.cse(cs.simplify(expr))
        out = super().constraint(name, expr, op, 0, soft, False)

        # NOTE: the line above already created the slack variables in each scenario, so
        # below we have to pass both soft=False and the expression with slack included.
        vars = self.variables
        pars = self.parameters
        expr_ = out[0]  # slack-relaxed expression in the form h(x,p)<=0 or g(x,p)==0
        op_: Literal["==", "<="] = "==" if op == "==" else "<="
        for i in range(self._starts):
            expr_i = _chained_substitute(
                expr_, vars, self._vars_i(i), pars, self._pars_i(i)
            )
            self._stacked_nlp.constraint(_n(name, i), expr_i, op_, 0, False, False)
        return out

    def minimize(self, objective: SymType) -> None:
        out = super().minimize(objective)

        vars = self.variables
        pars = self.parameters
        self._fs: list[SymType] = [
            _chained_substitute(objective, vars, self._vars_i(i), pars, self._pars_i(i))
            for i in range(self._starts)
        ]
        self._stacked_nlp.minimize(sum(self._fs))
        return out

    def init_solver(
        self,
        opts: Optional[dict[str, Any]] = None,
        solver: str = "ipopt",
        type: Optional[Literal["nlp", "conic"]] = None,
    ) -> None:
        out = super().init_solver(opts, solver, type)
        self._stacked_nlp.init_solver(opts, solver, type)
        return out

    def solve_multi(
        self,
        pars: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        return_all_sols: bool = False,
        return_stacked_sol: bool = False,
        **_,
    ) -> Union[Solution[SymType], list[Solution[SymType]]]:
        assert not (
            return_stacked_sol and return_all_sols
        ), "`return_all_sols` and `return_stacked_sol` can't be both true."
        if pars is not None:
            pars_iter = repeat(pars, self.starts) if isinstance(pars, dict) else pars
            pars = {
                _n(n, i): pars_i[n]
                for i, pars_i in enumerate(pars_iter)
                for n in pars_i.keys()
            }
        if vals0 is not None:
            v0_iter = repeat(vals0, self.starts) if isinstance(vals0, dict) else vals0
            vals0 = {
                _n(n, i): vals0_i[n]
                for i, vals0_i in enumerate(v0_iter)
                for n in vals0_i.keys()
            }
        multi_sol = self._stacked_nlp.solve(pars, vals0)
        if return_stacked_sol:
            return multi_sol

        x_ = self._x
        p_ = self._p
        lam_g_and_h_ = cs.vertcat(self._lam_g, self._lam_h)
        lam_lbx_and_ubx_ = cs.vertcat(self._lam_lbx, self._lam_ubx)
        vars_ = self.variables
        pars_ = self.parameters
        duals_ = self.dual_variables
        fs = [float(multi_sol.value(f)) for f in self._fs]

        all_vars = cs.vertcat(x_, lam_g_and_h_, lam_lbx_and_ubx_, p_)
        splits = np.cumsum(
            (0, x_.size1(), lam_g_and_h_.size1(), lam_lbx_and_ubx_.size1(), p_.size1())
        )
        solver_plugin = self.unwrapped._solver_plugin

        def get_ith_sol(idx: int) -> EagerSolution[SymType]:
            vals = {n: multi_sol.vals[_n(n, idx)] for n in vars_.keys()}
            dual_vals = {n: multi_sol.dual_vals[_n(n, idx)] for n in duals_.keys()}
            # get the value of p, and, lam g, lam h and lam lbx and lam ubx in a single
            # substitution to save some time
            all_vals = multi_sol.value(
                _chained_substitute(
                    all_vars,
                    vars_,
                    self._vars_i(idx),
                    pars_,
                    self._pars_i(idx),
                    duals_,
                    self._dual_vars_i(idx),
                )
            )
            x, lam_g_and_h, lam_lbx_and_ubx, p = cs.vertsplit(all_vals, splits)
            return EagerSolution(
                fs[idx],
                p_,
                p,
                x_,
                x,
                lam_g_and_h_,
                lam_g_and_h,
                lam_lbx_and_ubx_,
                lam_lbx_and_ubx,
                vars_,
                vals,
                duals_,
                dual_vals,
                multi_sol.stats,
                solver_plugin,
            )

        if return_all_sols:
            return [get_ith_sol(i) for i in range(self._starts)]
        return get_ith_sol(np.argmin(fs).item())

    def remove_variable_bounds(
        self,
        name: str,
        direction: Literal["lb", "ub", "both"],
        idx: Union[tuple[int, int], list[tuple[int, int]]] = None,
    ) -> None:
        idx = [idx] if isinstance(idx, tuple) else list(idx)
        super().remove_variable_bounds(name, direction, idx)
        for i in range(self._starts):
            self._stacked_nlp.remove_variable_bounds(_n(name, i), direction, idx)

    def remove_constraints(
        self, name: str, idx: Union[tuple[int, int], list[tuple[int, int]]] = None
    ) -> None:
        idx = [idx] if isinstance(idx, tuple) else list(idx)
        super().remove_constraints(name, idx)
        for i in range(self._starts):
            self._stacked_nlp.remove_constraints(_n(name, i), idx)


class ParallelMultistartNlp(MultistartNlp[SymType], Generic[SymType]):
    """A class that solves an NLP problem multiple times, with different initial
    starting conditions, via parallelization of the computations via :mod:`joblib`.

    Parameters
    ----------
    args, kwargs
        See inherited :meth:`csnlp.Nlp.__init__`.
    starts : int
        A positive integer for the number of multiple starting guesses to optimize.
    parallel_kwargs: dict, optional
        A dictionary with keyword arguments to be used to instantiate the parallel
        :class:`joblib.Parallel` backend. By default, ``None``, in which case the
        parallel backend is instantiated with default arguments.

    Raises
    ------
    ValueError
        Raises if the scenario number is invalid.
    """

    def __init__(
        self,
        *args: Any,
        starts: int,
        parallel_kwargs: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, starts=starts, **kwargs)
        self._parallel_kwargs = parallel_kwargs if parallel_kwargs is not None else {}
        self._parallel: Optional[Parallel] = None

    def initialize_parallel(self) -> None:
        """Initializes the parallel backend."""
        self._parallel = Parallel(**self._parallel_kwargs)
        self._parallel.__enter__()

    def terminate_parallel(self) -> None:
        """Terminates the parallel backend."""
        self._parallel.__exit__(None, None, None)

    def solve_multi(
        self,
        pars: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        return_all_sols: bool = False,
        **_,
    ) -> Union[Solution[SymType], list[Solution[SymType]]]:
        if self._solver is None:
            raise RuntimeError("Solver uninitialized.")
        if self._parallel is None:
            self.initialize_parallel()
        shared_kwargs = {
            "lbx": self._lbx.data,
            "ubx": self._ubx.data,
            "lbg": np.concatenate((np.zeros(self.ng), np.full(self.nh, -np.inf))),
            "ubg": 0,
        }
        pars_iter = (
            repeat(pars, self.starts)
            if pars is None or isinstance(pars, dict)
            else pars
        )
        vals0_iter = (
            repeat(vals0, self.starts)
            if vals0 is None or isinstance(vals0, dict)
            else vals0
        )
        kwargs = (
            self._process_pars_and_vals0(shared_kwargs.copy(), p, v0)
            for p, v0 in zip(pars_iter, vals0_iter)
        )
        sols: Iterator[dict[str, Any]] = self._parallel(
            delayed(_solve_and_get_stats)(self._solver, kw) for kw in kwargs
        )
        if return_all_sols:
            return [LazySolution.from_casadi_solution(sol, self) for sol in sols]
        best_sol = min(sols, key=lambda s: _cmp_key(s, self._solver_plugin))
        return LazySolution.from_casadi_solution(best_sol, self)

    def __getstate__(self, fullstate: bool = False) -> dict[str, Any]:
        # joblib.Parallel cannot be pickled or deepcopied
        state = super().__getstate__(fullstate)
        state.pop("_parallel", None)
        return state

    def __setstate__(self, state: Optional[dict[str, Any]]) -> None:
        if state is not None:
            self.__dict__.update(state)
        self._parallel = None


class MappedMultistartNlp(MultistartNlp[SymType], Generic[SymType]):
    """A class that solves an NLP problem multiple times, with different initial
    conditions, in parallel via :func:`casadi.Function.map` parallelization.

    See
    `this wiki <https://github.com/casadi/casadi/wiki/FAQ:-How-to-use-map%28%29-and-mapaccum%28%29-to-speed-up-calculations%3F>`_
    for more details on how to enable parallelization via mapping in CasADi.

    Parameters
    ----------
    args, kwargs
        See inherited :meth:`csnlp.Nlp.__init__`.
    starts : int
        A positive integer for the number of multiple starting guesses to optimize.
    parallelization : "serial", "unroll", "inline", "thread", "openmp"
        The type of parallelization to use (see :func:`casadi.Function.map`). By
        default, ``"serial"`` is selected.
    max_num_threads : int, optional
        Maximum number of threads to use in parallelization; if ``None``, the number of
        threads is equal to the number of starts.

    Raises
    ------
    ValueError
        Raises if the scenario number is invalid.
    """

    def __init__(
        self,
        *args: Any,
        starts: int,
        parallelization: Literal[
            "serial", "unroll", "inline", "thread", "openmp"
        ] = "serial",
        max_num_threads: Optional[int] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, starts=starts, **kwargs)
        self._mapped_solver: Optional[MemorizedFunc] = None
        self._parallelization = parallelization
        self._max_num_threads = max_num_threads

    def init_solver(self, *args: Any, **kwargs: Any) -> None:
        super().init_solver(*args, **kwargs)
        solver: cs.Function = self._solver.func
        mapped_solver = solver.map(
            self._starts, self._parallelization, self._max_num_threads or self._starts
        )
        self._mapped_solver = self._cache.cache(mapped_solver)

    def solve_multi(
        self,
        pars: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        return_all_sols: bool = False,
        _return_mapped_sol: bool = False,
        **_,
    ) -> Union[Solution[SymType], list[Solution[SymType]]]:
        assert not (
            _return_mapped_sol and return_all_sols
        ), "`return_all_sols` and `_return_mapped_sol` can't be both true."
        if self._mapped_solver is None:
            raise RuntimeError("Solver uninitialized.")
        pars_iter = (
            repeat(pars, self.starts)
            if pars is None or isinstance(pars, dict)
            else pars
        )
        vals0_iter = (
            repeat(vals0, self.starts)
            if vals0 is None or isinstance(vals0, dict)
            else vals0
        )
        x0s = []
        ps = []
        default = cs.DM.zeros(0, 1)
        for p, v0 in zip(pars_iter, vals0_iter):
            kwargs = self._process_pars_and_vals0({}, p, v0)
            x0s.append(kwargs.get("x0", default))
            ps.append(kwargs.get("p", default))
        single_kwargs = {
            "x0": cs.hcat(x0s),
            "p": cs.hcat(ps),
            "lbx": self._lbx.data,
            "ubx": self._ubx.data,
            "lbg": np.concatenate((np.zeros(self.ng), np.full(self.nh, -np.inf))),
            "ubg": 0,
        }
        single_sol: dict[str, cs.DM] = self._mapped_solver(**single_kwargs)

        # NOTE: the mapped solver does not return the stats, so we have to use the
        # original solver to get them - this likely returns the stats of the last
        # run of the mapped solver, but it's the best we can do up to this point
        stats = self._solver.func.stats()

        # if the user wants the mapped solution, return it, though it's not a Solution
        # and it's mostly for debugging
        if _return_mapped_sol:
            single_sol["p"] = single_kwargs["p"]
            single_sol["stats"] = stats
            return single_sol  # NOTE: this is NOT a Solution object

        # convert the mapped solution to a list of Solution objects
        if return_all_sols:
            sols = []
            for i, p in enumerate(ps):
                sol_i = {k: v[:, i] for k, v in single_sol.items()}
                sol_i["p"] = ps[i]
                sol_i["stats"] = stats.copy()
                sols.append(LazySolution.from_casadi_solution(sol_i, self))
            return sols

        # just return the best, as we cannot use stats to exclude failed solutions
        i = np.argmin(single_sol["f"])
        sol = {k: v[:, i] for k, v in single_sol.items()}
        sol["p"] = ps[i]
        sol["stats"] = stats
        return LazySolution.from_casadi_solution(sol, self)



from typing import Any, Literal, Optional, TypeVar

import casadi as cs
import numpy as np
import numpy.typing as npt
from joblib import Memory
from joblib.memory import MemorizedFunc

from ..core.solutions import LazySolution, Solution, subsevalf
from .constraints import HasConstraints

SymType = TypeVar("SymType", cs.SX, cs.MX)


def _solve_and_get_stats(
    solver: MemorizedFunc, kwargs: dict[str, npt.ArrayLike]
) -> dict[str, Any]:
    """Internal utility to simultaneously run the solver and get its stats."""
    sol = solver(**kwargs)
    sol["p"] = kwargs["p"]  # add to solution the parameters for which it was computed
    sol["stats"] = solver.func.stats()  # add to solution the solver stats
    # NOTE: in case of failure in retrieving the stats
    # try:
    #     stats = solver.func.stats()
    # except RuntimeError:
    #     stats = {"success": False}
    # sol["stats"] = stats  # add to solution the solver stats
    return sol


class HasObjective(HasConstraints[SymType]):
    r"""Class for creating an NLP problem with parameters, variables, constraints and an
    objective. It builds on top of :class:`HasConstraints`, which handles parameters,
    variables and constraints.

    Parameters
    ----------
    sym_type : {"SX", "MX"}, optional
        The CasADi symbolic variable type to use in the NLP, by default ``"SX"``.
    remove_redundant_x_bounds : bool, optional
        If ``True``, then redundant entries in :meth:`lbx` and :meth:`ubx` are removed
        when properties :meth:`h_lbx` and :meth:`h_ubx` are called. See these two
        properties for more details. By default, ``True``.
    cache : joblib.Memory, optional
        Optional cache to avoid computing the same exact NLP more than once. By default,
        no caching occurs.
    name : str, optional
        Name of the NLP scheme. If ``None``, it is automatically assigned.

    Notes
    -----
    Constraints are internally handled in their canonical form, i.e., :math:`g(x,p) = 0`
    and :math:`h(x,p) \leq 0`. The objective :math:`f(x,p)` is always a scalar function
    to be minimized.
    """

    def __init__(
        self,
        sym_type: Literal["SX", "MX"] = "SX",
        remove_redundant_x_bounds: bool = True,
        cache: Memory = None,
        name: Optional[str] = None,
    ) -> None:
        super().__init__(sym_type, remove_redundant_x_bounds)
        self.name = name
        self._f: Optional[SymType] = None
        self._solver: Optional[MemorizedFunc] = None
        self._solver_opts: dict[str, Any] = {}
        self._cache = cache if cache is not None else Memory(None)
        self._failures = 0

    @property
    def f(self) -> Optional[SymType]:
        """Gets the objective of the NLP scheme, which is ``None`` if not set previously
        set via the :meth:`minimize` method."""
        return self._f

    @property
    def solver(self) -> Optional[cs.Function]:
        """Gets the NLP optimization solver. Can be ``None``, if the solver is not set
        with method :meth:`init_solver`."""
        return self._solver.func if self._solver is not None else None

    @property
    def solver_opts(self) -> dict[str, Any]:
        """Gets the NLP optimization solver options. The dict is empty, if the solver
        options are not set with method :meth:`init_solver`."""
        return self._solver_opts

    @property
    def failures(self) -> int:
        """Gets the cumulative number of failures of the NLP solver."""
        return self._failures

    def init_solver(
        self,
        opts: Optional[dict[str, Any]] = None,
        solver: str = "ipopt",
        type: Optional[Literal["nlp", "conic"]] = None,
    ) -> None:
        """Initializes the solver for this NLP with the given options.

        Parameters
        ----------
        opts : dict[str, Any], optional
            Options to be passed to the CasADi interface to the solver. Must not contain
            the ``"discrete"`` key, which is automatically set based on the variables'
            domains. By default, ``None``.
        solver : str, optional
            Type of solver to instantiate. For example, ``"ipopt"`` and ``"sqpmethod"``
            trigger the instantiation of an NLP problem, while, e.g., ``"qrqp"``,
            ``"osqp"``, and ``"qpoases"`` of a conic one. However, the solver type can
            be overruled with the ``type`` argument. By default, ``"ipopt"`` is
            selected.
        type : "nlp", "conic", optional
            Type of problem to instantiate. If ``"nlp"``, then the problem is forced to
            be instantiated with :func:`casadi.nlpsol`. If ``"conic"``, then
            :func:`casadi.qpsol` is forced instead. If ``None``, then the problem type
            is selected automatically.

        Raises
        ------
        ValueError
            Raises if the given problem type is not recognized, or if the ``opts`` dict
            contains the
        RuntimeError
            Raises if the type of the problem cannot be inferred automatically (when the
            solver supports both conic and NLPs), if the specified solver plugin cannot
            be found, and if the objective has not yet been specified with
            :meth:`minimize`.
        """
        has_conic = cs.has_conic(solver)
        has_nlpsol = cs.has_nlpsol(solver)
        auto_type = type is None
        if has_conic and has_nlpsol and auto_type:
            raise RuntimeError(
                f"`{solver}` supports both conic- and nlp-type problems, so the problem"
                " type cannot be inferred automatically. Please, provide also argument"
                " 'type' to the method."
            )

        if type == "conic" or (auto_type and has_conic):
            func = cs.qpsol
        elif type == "nlp" or (auto_type and has_nlpsol):
            func = cs.nlpsol
        elif type not in (None, "nlp", "conic"):
            raise ValueError(f"unknown problem type: '{type}'")
        else:
            raise RuntimeError(f"'{solver}' plugin not found in either conic or nlp.")

        if self._f is None:
            raise RuntimeError("NLP objective not set.")

        opts = {} if opts is None else opts.copy()
        if "discrete" in opts:
            raise ValueError("The 'discrete' key is reserved for the variable domains.")
        if self.has_discrete:
            disc = self.discrete
            opts["discrete"] = [disc.item()] if disc.size == 1 else disc  # bugfix
        con = cs.vertcat(self._g, self._h)
        problem = {"x": self._x, "p": self._p, "g": con, "f": self._f}
        solver_func = func(f"solver_{solver}_{self.name}", solver, problem, opts)

        self._solver = self._cache.cache(solver_func)
        opts.pop("discrete", None)
        self._solver_opts = opts
        self._solver_plugin = solver
        self._solver_type = type

    def refresh_solver(self) -> None:
        """Refresh and resets the internal solver function (with the same options, if
        previously set)."""
        if self._solver is not None:
            self.init_solver(self._solver_opts, self._solver_plugin, self._solver_type)

    def minimize(self, objective: SymType) -> None:
        """Sets the objective function to be minimized.

        Parameters
        ----------
        objective : casadi SX or MX
            A Symbolic variable dependent only on the NLP variables and parameters that
            needs to be minimized.

        Raises
        ------
        ValueError
            Raises if the objective is not scalar.
        """
        if not objective.is_scalar():
            raise ValueError("Objective must be scalar.")
        self._f = objective
        self.refresh_solver()

    def parameter(self, *args, **kwargs):
        out = super().parameter(*args, **kwargs)
        self.refresh_solver()
        return out

    def variable(self, *args, **kwargs):
        out = super().variable(*args, **kwargs)
        self.refresh_solver()
        return out

    def constraint(self, *args, **kwargs):
        out = super().constraint(*args, **kwargs)
        self.refresh_solver()
        return out

    def solve(
        self,
        pars: Optional[dict[str, npt.ArrayLike]] = None,
        vals0: Optional[dict[str, npt.ArrayLike]] = None,
    ) -> Solution[SymType]:
        """Solves the NLP optimization problem.

        Parameters
        ----------
        pars : dict[str, array_like], optional
            Dictionary or structure containing, for each parameter in the NLP scheme,
            the corresponding numerical value. Can be `None` if no parameters are
            present.
        vals0 : dict[str, array_like], optional
            Dictionary or structure containing, for each variable in the NLP scheme, the
            corresponding initial guess. By default, initial guesses are not passed to
            the solver.

        Returns
        -------
        sol : Solution
            A solution object containing all the information.

        Raises
        ------
        RuntimeError
            Raises if the solver is un-initialized (see :meth:`init_solver`); or if not
            all the parameters are not provided with a numerical value.
        """
        if self._solver is None:
            raise RuntimeError("Solver uninitialized.")
        kwargs = self._process_pars_and_vals0(
            {
                "lbx": self._lbx.data,
                "ubx": self._ubx.data,
                "lbg": np.concatenate((np.zeros(self.ng), np.full(self.nh, -np.inf))),
                "ubg": 0,
            },
            pars,
            vals0,
        )
        sol_with_stats = _solve_and_get_stats(self._solver, kwargs)
        solution = LazySolution.from_casadi_solution(sol_with_stats, self)
        self._failures += not solution.success
        return solution

    def _process_pars_and_vals0(
        self,
        kwargs: dict[str, npt.ArrayLike],
        pars: Optional[dict[str, npt.ArrayLike]],
        vals0: Optional[dict[str, npt.ArrayLike]],
    ) -> dict[str, npt.ArrayLike]:
        """Internal utility to convert pars and initial-val dicts to solver kwargs."""
        if self._pars:
            if pars is None:
                pars = {}
            if parsdiff := self._pars.keys() - pars.keys():
                raise RuntimeError(
                    "Trying to solve the NLP with unspecified parameters: "
                    + ", ".join(parsdiff)
                    + "."
                )
            kwargs["p"] = subsevalf(self._p, self._pars, pars)
        else:
            kwargs["p"] = cs.DM()
        if vals0 is not None:
            if vals0diff := self._vars.keys() - vals0.keys():
                vals0.update({v: 0 for v in vals0diff})
            kwargs["x0"] = subsevalf(self._x, self._vars, vals0)
        return kwargs



import pickle
import unittest
from itertools import product
from math import ceil, floor
from random import choice
from unittest.mock import Mock

import casadi as cs
import numpy as np
from parameterized import parameterized

from csnlp import Nlp
from csnlp.core.solutions import subsevalf
from csnlp.wrappers import Mpc, PwaMpc, PwaRegion
from csnlp.wrappers import ScenarioBasedMpc as SCMPC
from csnlp.wrappers.mpc.scenario_based_mpc import _n

OPTS = {
    "expand": True,
    "print_time": False,
    "ipopt": {
        "max_iter": 500,
        "tol": 1e-7,
        "barrier_tol_factor": 5,
        "sb": "yes",
        # for debugging
        "print_level": 0,
        "print_user_options": "no",
        "print_options_documentation": "no",
    },
}


class TestMpc(unittest.TestCase):
    def test_init__raises__with_invalid_args(self):
        nlp = Nlp()
        with self.assertRaisesRegex(ValueError, "Invalid shooting method."):
            Mpc(nlp, shooting="ciao", prediction_horizon=1)
        with self.assertRaisesRegex(
            ValueError, "Prediction horizon must be positive and > 0."
        ):
            Mpc(nlp, prediction_horizon=0)
        with self.assertRaisesRegex(
            ValueError, "Control horizon must be positive and > 0."
        ):
            Mpc(nlp, prediction_horizon=10, control_horizon=-2)

    def test_init__initializes_control_horizon_properly(self):
        N = 10
        nlp = Nlp()
        mpc1 = Mpc[cs.SX](nlp, prediction_horizon=N)
        mpc2 = Mpc[cs.SX](nlp, prediction_horizon=N, control_horizon=N * 2)
        self.assertEqual(mpc1.control_horizon, N)
        self.assertEqual(mpc2.control_horizon, N * 2)

    @parameterized.expand([("single",), ("multi",)])
    def test_state__constructs_state_correctly(self, shooting: str):
        N = 10
        nlp = Nlp(sym_type="MX")
        mpc = Mpc[cs.MX](nlp=nlp, prediction_horizon=N, shooting=shooting)
        x1, x1_0 = mpc.state("x1", 2)
        if shooting == "multi":
            self.assertEqual(x1.shape, (2, N + 1))
            self.assertEqual(x1.shape, mpc.states["x1"].shape)
            self.assertEqual(mpc.constraints["x1_0"].shape, (2, 1))
        else:
            self.assertIsNone(x1)
        self.assertEqual(x1_0.shape, (2, 1))
        self.assertEqual(x1_0.shape, mpc.initial_states["x1_0"].shape)
        self.assertEqual(mpc.ns, x1_0.shape[0])
        x2, x2_0 = mpc.state("x2", 1)
        if shooting == "multi":
            self.assertEqual(x2.shape, (1, N + 1))
            self.assertEqual(x2.shape, mpc.states["x2"].shape)
            self.assertEqual(mpc.constraints["x2_0"].shape, (1, 1))
        else:
            self.assertIsNone(x2)
        self.assertEqual(x2_0.shape, (1, 1))
        self.assertEqual(x2_0.shape, mpc.initial_states["x2_0"].shape)
        self.assertEqual(mpc.ns, x1_0.shape[0] + x2_0.shape[0])

    @parameterized.expand(product((False, True), (False, True)))
    def test_state__removes_bounds_properly(self, initial: bool, terminal: bool):
        N = 10
        nlp = Nlp(sym_type="MX")
        mpc = Mpc[cs.MX](nlp=nlp, prediction_horizon=N, shooting="multi")

        shape = (2, N + 1)
        nlp.variable = Mock(return_value=cs.MX.sym("x", *shape))
        mpc.state("x", 2, lb=0, ub=1, bound_initial=initial, bound_terminal=terminal)

        lb, ub = np.full(shape, 0.0), np.full(shape, 1.0)
        if not initial:
            lb[:, 0], ub[:, 0] = -np.inf, np.inf
        if not terminal:
            lb[:, -1], ub[:, -1] = -np.inf, np.inf
        nlp.variable.assert_called_once()
        lb_actual, ub_actual = nlp.variable.call_args[0][-2:]
        np.testing.assert_array_equal(lb_actual, lb)
        np.testing.assert_array_equal(ub_actual, ub)

    @parameterized.expand([(0,), (1,), (2,)])
    def test_state__raises__in_singleshooting_with_state_bounds(self, i: int):
        nlp = Nlp(sym_type="MX")
        mpc = Mpc[cs.MX](nlp=nlp, prediction_horizon=10, shooting="single")
        with self.assertRaises(RuntimeError):
            if i == 0:
                mpc.state("x1", 2, lb=0)
            elif i == 1:
                mpc.state("x1", 2, ub=1)
            else:
                mpc.state("x1", 2, lb=0, ub=1)

    @parameterized.expand(product((1, 2), (1, 3)))
    def test_action__constructs_action_correctly(self, divider: int, space: int):
        Np = 10
        Nc = Np // divider
        nlp = Nlp(sym_type="SX")
        mpc = Mpc[cs.SX](
            nlp=nlp, prediction_horizon=Np, control_horizon=Nc, input_spacing=space
        )
        u1, u1_exp = mpc.action("u1", 2)
        self.assertEqual(u1.shape, (2, ceil(Nc / space)))
        self.assertEqual(u1.shape, mpc.actions["u1"].shape)
        self.assertEqual(u1_exp.shape, (2, Np))
        self.assertEqual(u1_exp.shape, mpc.actions_expanded["u1"].shape)
        self.assertEqual(mpc.na, u1.shape[0])
        u2, u2_exp = mpc.action("u2", 1)
        self.assertEqual(u2.shape, (1, ceil(Nc / space)))
        self.assertEqual(u2.shape, mpc.actions["u2"].shape)
        self.assertEqual(u2_exp.shape, (1, Np))
        self.assertEqual(u2_exp.shape, mpc.actions_expanded["u2"].shape)
        for i in range(Nc):
            self.assertTrue(cs.is_equal(u1[:, floor(i / space)], u1_exp[:, i]))
            self.assertTrue(cs.is_equal(u2[:, floor(i / space)], u2_exp[:, i]))
        for i in range(Nc - 1, Np):
            self.assertTrue(cs.is_equal(u1[:, -1], u1_exp[:, i]))
            self.assertTrue(cs.is_equal(u2[:, -1], u2_exp[:, i]))
        self.assertEqual(mpc.na, u1.shape[0] + u2.shape[0])

    def test_constraint__constructs_slack_correctly(self):
        nlp = Nlp(sym_type="MX")
        mpc = Mpc[cs.MX](nlp=nlp, prediction_horizon=10)
        x, _ = mpc.state("x1", 3)
        _, _, slack = mpc.constraint("c0", x, ">=", 5, soft=True)
        self.assertIn(slack.name(), mpc.slacks)
        self.assertIn(slack.name(), mpc.slacks)
        self.assertEqual(slack.shape, mpc.slacks[slack.name()].shape)
        self.assertEqual(mpc.nslacks, x.shape[0])
        mpc.constraint("c1", x, "<=", 10, soft=False)
        self.assertEqual(mpc.nslacks, x.shape[0])

    def test_disturbance__constructs_disturbance_correctly(self):
        N = 10
        nlp = Nlp(sym_type="MX")
        mpc = Mpc[cs.MX](nlp=nlp, prediction_horizon=N)
        d1 = mpc.disturbance("d1", 2)
        self.assertEqual(d1.shape, (2, N))
        self.assertEqual(d1.shape, mpc.disturbances["d1"].shape)
        self.assertEqual(mpc.nd, 2)
        d2 = mpc.disturbance("d2", 20)
        self.assertEqual(d2.shape, (20, N))
        self.assertEqual(d2.shape, mpc.disturbances["d2"].shape)
        self.assertEqual(mpc.nd, 22)

    def test_nonlinear_dynamics__raises__if_dynamics_already_set(self):
        nlp = Nlp(sym_type="MX")
        mpc = Mpc(nlp=nlp, prediction_horizon=10)
        mpc._dynamics_already_set = True
        with self.assertRaises(RuntimeError):
            mpc.set_nonlinear_dynamics(object())

    def test_nonlinear_dynamics__raises__if_dynamics_arguments_are_invalid(self):
        x1 = cs.SX.sym("x1", 2)
        x2 = cs.SX.sym("x2", 3)
        x = cs.vertcat(x1, x2)
        u1 = cs.SX.sym("u1", 3)
        u2 = cs.SX.sym("u2", 1)
        u = cs.vertcat(u1, u2)
        d = cs.SX.sym("d")
        p = cs.SX.sym("p")
        x_next = x + cs.vertcat(u, u2)
        F1 = cs.Function("F", [x], [x_next], {"allow_free": True})
        F2 = cs.Function("F", [x, u, d, p], [x_next + d + p])
        nlp = Nlp(sym_type="MX")
        mpc = Mpc(nlp=nlp, prediction_horizon=10, control_horizon=5)
        for F in (F1, F2):
            with self.assertRaises(ValueError):
                mpc.set_nonlinear_dynamics(F)

    @parameterized.expand([(0,), (1,)])
    def test_nonlinear_dynamics__in_multishooting__creates_dynamics_eq_constraints(
        self, i: int
    ):
        N = 10
        nlp = Nlp(sym_type="SX")
        mpc = Mpc[cs.SX](
            nlp=nlp, prediction_horizon=N, control_horizon=N // 2, shooting="multi"
        )
        x1, _ = mpc.state("x1", 2)
        x2, _ = mpc.state("x2", 3)
        u1, _ = mpc.action("u1", 3)
        u2, _ = mpc.action("u2", 1)
        x = cs.vertcat(x1[:, 0], x2[:, 0])
        u = cs.vertcat(u1[:, 0], u2[:, 0])
        x_next = x + cs.vertcat(u, u2[:, 0])
        if i == 0:
            F = cs.Function("F", [x, u], [x_next], ["x", "u"], ["x+"])
        else:
            d = mpc.disturbance("d")
            x_next += d[:, 0]
            F = cs.Function("F", [x, u, d], [x_next], ["x", "u", "d"], ["x+"])
        mpc.set_nonlinear_dynamics(F)
        self.assertIn("dyn", mpc.constraints.keys())
        self.assertEqual(mpc.ng, (1 + N) * 5)

    @parameterized.expand([(0,), (1,)])
    def test_nonlinear_dynamics__in_singleshooting__creates_states(self, i: int):
        N = 10
        nlp = Nlp(sym_type="SX")
        mpc = Mpc[cs.SX](
            nlp=nlp, prediction_horizon=N, control_horizon=N // 2, shooting="single"
        )
        mpc.state("x1", 2)
        x1 = cs.SX.sym("x1", 2)
        mpc.state("x2", 3)
        x2 = cs.SX.sym("x2", 3)
        u1, _ = mpc.action("u1", 3)
        u2, _ = mpc.action("u2", 1)
        x = cs.vertcat(x1[:, 0], x2[:, 0])
        u = cs.vertcat(u1[:, 0], u2[:, 0])
        x_next = x + cs.vertcat(u, u2[:, 0])
        if i == 0:
            F = cs.Function("F", [x, u], [x_next], ["x", "u"], ["x+"])
        else:
            d = mpc.disturbance("d")
            x_next += d[:, 0]
            F = cs.Function("F", [x, u, d], [x_next], ["x", "u", "d"], ["x+"])
        mpc.set_nonlinear_dynamics(F)
        for k in range(N):
            self.assertNotIn(f"dyn_{k}", mpc.constraints.keys())
        self.assertIn("x1", mpc.states.keys())
        self.assertIn("x2", mpc.states.keys())
        self.assertEqual(mpc.states["x1"].shape, (2, N + 1))
        self.assertEqual(mpc.states["x2"].shape, (3, N + 1))

    def test_affine_dynamics__raises__if_dynamics_already_set(self):
        nlp = Nlp(sym_type="MX")
        mpc = Mpc(nlp=nlp, prediction_horizon=10)
        mpc._dynamics_already_set = True
        with self.assertRaises(RuntimeError):
            mpc.set_affine_dynamics(object(), object())

    @parameterized.expand([("A",), ("B",), ("D",), ("c",)])
    def test_affine_dynamics__raises__if_matrices_have_wrong_shapes(self, wrong_mat):
        ns, na, nd = np.random.randint(4, 20, size=3)
        nlp = Nlp(sym_type="MX")
        mpc = Mpc(nlp=nlp, prediction_horizon=10)
        mpc.state("x", ns)
        mpc.action("u", na)
        mpc.disturbance("d", nd)
        shapes = {"A": (ns, ns), "B": (ns, na), "D": (ns, nd), "c": (ns,)}

        wrong_shapes = shapes.copy()
        wrong_shapes[wrong_mat] = np.add(shapes[wrong_mat], choice((1, -1)))

        A = np.random.randn(*wrong_shapes["A"])
        B = np.random.randn(*wrong_shapes["B"])
        D = np.random.randn(*wrong_shapes["D"])
        c = np.random.randn(*wrong_shapes["c"])
        with self.assertRaisesRegex(ValueError, f"{wrong_mat} must have shape"):
            mpc.set_affine_dynamics(A, B, D, c)

    @parameterized.expand([(0,), (6,)])
    def test_affine_dynamics__raises__when_disturbances_are_misspecified(self, nd: int):
        ns, na = np.random.randint(4, 20, size=2)
        nlp = Nlp(sym_type="MX")
        mpc = Mpc(nlp=nlp, prediction_horizon=10)
        mpc.state("x", ns)
        mpc.action("u", na)
        A = np.random.randn(ns, ns)
        B = np.random.randn(ns, na)

        if nd == 0:
            D = np.random.randn(ns, 123)
            with self.assertRaisesRegex(ValueError, "Expected D to be `None`"):
                mpc.set_affine_dynamics(A, B, D)
        else:
            mpc.disturbance("d", nd)
            D = None
            with self.assertRaisesRegex(ValueError, "D must be provided"):
                mpc.set_affine_dynamics(A, B, D)

    @parameterized.expand(product((False, True), (False, True)))
    def test_affine_dynamics__in_multishooting__creates_dynamics_eq_constraints(
        self, include_d: bool, include_c: bool
    ):
        ns, na, nd, N = np.random.randint(4, 20, size=4)
        nlp = Nlp(sym_type="MX")
        mpc = Mpc(nlp=nlp, prediction_horizon=N)
        mpc.state("x", ns)
        mpc.action("u", na)
        A = np.random.randn(ns, ns)
        B = np.random.randn(ns, na)
        if include_d:
            mpc.disturbance("d", nd)
            D = np.random.randn(ns, nd)
        else:
            D = None
        c = np.random.randn(ns) if include_c else None

        mpc.set_affine_dynamics(A, B, D, c)

        self.assertIn("dyn", mpc.constraints.keys())
        self.assertEqual(mpc.ng, (N + 1) * ns)

    @parameterized.expand(product((False, True), (False, True)))
    def test_affine_dynamics__in_singleshooting__creates_states(
        self, include_d: bool, include_c: bool
    ):
        ns, na, nd, N = np.random.randint(4, 20, size=4)
        nlp = Nlp(sym_type="MX")
        mpc = Mpc(nlp=nlp, prediction_horizon=N)
        mpc.state("x", ns)
        mpc.action("u", na)
        A = np.random.randn(ns, ns)
        B = np.random.randn(ns, na)
        if include_d:
            mpc.disturbance("d", nd)
            D = np.random.randn(ns, nd)
        else:
            D = None
        c = np.random.randn(ns) if include_c else None

        mpc.set_affine_dynamics(A, B, D, c)

        for k in range(N):
            self.assertNotIn(f"dyn_{k}", mpc.constraints.keys())
        self.assertIn("x", mpc.states.keys())
        self.assertEqual(mpc.states["x"].shape, (ns, N + 1))

    @parameterized.expand([("SX",), ("MX",)])
    def test_can_be_pickled(self, sym_type: str):
        N = 10
        mpc = Mpc(
            nlp=Nlp(sym_type=sym_type), prediction_horizon=N, control_horizon=N // 2
        )
        mpc.state("x1", 2)
        mpc.state("x2", 3)
        mpc.action("u1", 3)
        mpc.action("u2", 1)

        mpc2 = pickle.loads(pickle.dumps(mpc))

        self.assertIn(repr(mpc), repr(mpc2))


class TestPwaMpc(unittest.TestCase):
    def test_pwa_dynamics__raises__if_dynamics_already_set(self):
        nlp = Nlp(sym_type="MX")
        mpc = PwaMpc(nlp=nlp, prediction_horizon=10)
        mpc._dynamics_already_set = True
        with self.assertRaises(RuntimeError):
            mpc.set_pwa_dynamics(object(), object(), object())

    @parameterized.expand(product(("SX", "MX"), range(3), (False, True)))
    def test_pwa_dynamics__raises__if_lower_or_upper_bounds_are_set(
        self, sym_type: str, set_lower: bool, set_state: bool
    ):
        mpc = PwaMpc(Nlp(sym_type=sym_type), prediction_horizon=2)
        kwargs = {"lb" if set_lower else "ub": 0} if set_state else {}
        mpc.state("x", 1, **kwargs)
        kwargs = {"lb" if set_lower else "ub": 0} if not set_state else {}
        mpc.action("u", 1, **kwargs)

        regex = (
            "Cannot set lower and upper bounds on the "
            f"{'states' if set_state else 'actions'} in PWA systems; use "
            "arguments `D` and `E` of `set_pwa_dyanmics` instead."
        )
        with self.assertRaisesRegex(RuntimeError, regex):
            mpc.set_pwa_dynamics(None, None, None)

    def test_pwa_linear_dynamics__raises__if_matrices_have_wrong_shapes(self):
        nx, nu, n_ineq_r, n_ineq_xu = np.random.randint(4, 20, size=4)
        A = np.random.randn(nx, nx)
        B = np.random.randn(nx, nu)
        c = np.random.randn(nx)
        S = np.random.randn(n_ineq_r, nx + nu)
        T = np.random.randn(n_ineq_r)
        D = np.random.randn(n_ineq_xu, nx + nu)
        E = np.random.randn(n_ineq_xu)
        args = {"A": A, "B": B, "c": c, "S": S, "T": T, "D": D, "E": E}

        for key in args:
            shape = args[key].shape
            new_args = args.copy()
            new_args[key] = np.random.randn(*np.add(shape, choice((1, -1))))

            region = PwaRegion(
                new_args["A"],
                new_args["B"],
                new_args["c"],
                new_args["S"],
                new_args["T"],
            )
            nlp = Nlp(sym_type="MX")
            mpc = PwaMpc(nlp=nlp, prediction_horizon=3, control_horizon=5)
            mpc.state("x", nx)
            mpc.action("u", nu)
            regions = [region]
            with self.assertRaises(ValueError):
                mpc.set_pwa_dynamics(regions, new_args["D"], new_args["E"])

    def test_pwa_dynamics__in_multishooting__creates_dynamics_eq_constraints(self):
        tau, k1, k2, d, m = 0.5, 10, 1, 4, 10
        A1 = np.array([[1, tau], [-((tau * 2 * k1) / m), 1 - (tau * d) / m]])
        A2 = np.array([[1, tau], [-((tau * 2 * k2) / m), 1 - (tau * d) / m]])
        B1 = B2 = np.array([[0], [tau / m]])
        x_bnd = (5, 5)
        u_bnd = 20
        pwa_regions = (
            PwaRegion(A1, B1, np.zeros(2), np.array([[1, 0, 0]]), np.zeros(1)),
            PwaRegion(A2, B2, np.zeros(2), np.array([[-1, 0, 0]]), np.zeros(1)),
        )
        D1 = np.array([[1, 0], [-1, 0], [0, 1], [0, -1]])
        E1 = np.array([x_bnd[0], x_bnd[0], x_bnd[1], x_bnd[1]])
        D2 = np.array([[1], [-1]])
        E2 = np.array([u_bnd, u_bnd])
        D = cs.diagcat(D1, D2).sparse()
        E = np.concatenate((E1, E2))
        N, ns, na = 4, 2, 1
        mpc = PwaMpc(Nlp(), prediction_horizon=N, shooting="multi")
        mpc.state("x", ns)
        mpc.action("u", na)
        mpc.set_pwa_dynamics(pwa_regions, D, E, parallelization="inline")

        nr = len(pwa_regions)
        constraints = {
            "delta_sum": (1, N),
            "z_ub": (ns * nr, N),
            "z_lb": (ns * nr, N),
            "z_x_ub": (ns * nr, N),
            "z_x_lb": (ns * nr, N),
            "region": (nr, N),
        }
        for con, shape in constraints.items():
            self.assertIn(con, mpc.constraints)
            self.assertEqual(mpc.constraints[con].shape, shape, msg=con)

    def test_pwa_dynamics__in_singleshooting__creates_states(self):
        tau, k1, k2, d, m = 0.5, 10, 1, 4, 10
        A1 = np.array([[1, tau], [-((tau * 2 * k1) / m), 1 - (tau * d) / m]])
        A2 = np.array([[1, tau], [-((tau * 2 * k2) / m), 1 - (tau * d) / m]])
        B1 = B2 = np.array([[0], [tau / m]])
        x_bnd = (5, 5)
        u_bnd = 20
        pwa_regions = (
            PwaRegion(A1, B1, np.zeros(2), np.array([[1, 0, 0]]), np.zeros(1)),
            PwaRegion(A2, B2, np.zeros(2), np.array([[-1, 0, 0]]), np.zeros(1)),
        )
        D1 = np.array([[1, 0], [-1, 0], [0, 1], [0, -1]])
        E1 = np.array([x_bnd[0], x_bnd[0], x_bnd[1], x_bnd[1]])
        D2 = np.array([[1], [-1]])
        E2 = np.array([u_bnd, u_bnd])
        D = cs.diagcat(D1, D2).sparse()
        E = np.concatenate((E1, E2))
        N, ns, na = 4, 2, 1
        mpc = PwaMpc(Nlp(), prediction_horizon=N, shooting="single")
        mpc.state("x", ns)
        mpc.action("u", na)
        mpc.set_pwa_dynamics(pwa_regions, D, E, parallelization="inline")

        self.assertIn("x", mpc.states)
        self.assertEqual(mpc.states["x"].shape, (ns, N + 1))


class TestScenarioBasedMpc(unittest.TestCase):
    @parameterized.expand([("multi",), ("single",)])
    def test_state__creates_as_many_states_as_scenarios(self, shooting: str):
        K, N, ns = np.random.randint(2, 20, size=3)
        ny = ns // 2
        nx = ns - ny
        scmpc = SCMPC[cs.MX](Nlp(sym_type="MX"), K, N, shooting=shooting)
        x, xs, _ = scmpc.state("x", nx, bound_initial=False, bound_terminal=False)
        y, ys, _ = scmpc.state("y", ny, bound_initial=False, bound_terminal=False)

        self.assertEqual(len(xs), K)
        self.assertEqual(len(ys), K)
        self.assertEqual(scmpc.ns, ns)
        self.assertEqual(scmpc.ns_all, ns * K)

        xshape = (nx, N + 1)
        yshape = (ny, N + 1)
        if shooting == "multi":
            self.assertTrue(all(x.shape == x_i.shape == xshape for x_i in xs))
            self.assertTrue(all(y.shape == y_i.shape == yshape for y_i in ys))
        else:
            self.assertEqual(x.shape, xshape)
            self.assertEqual(y.shape, yshape)
            self.assertTrue(all(x_i is None for x_i in xs))
            self.assertTrue(all(y_i is None for y_i in ys))

    def test_disturbance__creates_as_many_disturbances_as_scenarios(self):
        K = np.random.randint(2, 20)
        scmpc = SCMPC[cs.MX](Nlp(sym_type="MX"), K, 10)
        size = 5
        d, ds = scmpc.disturbance("d", size)
        self.assertEqual(len(ds), K)
        self.assertEqual(scmpc.nd, size)
        self.assertEqual(scmpc.nd_all, size * K)
        self.assertEqual(d.shape, ds[0].shape)

    @parameterized.expand([("SX",), ("MX",)])
    def test_dynamics__in_multishooting__creates_dynamics_eq_constraints(self, sym_tpe):
        nx, nu, nd, N, K = np.random.randint(4, 20, size=5)
        x, u, d = cs.SX.sym("x", nx), cs.SX.sym("u", nu), cs.SX.sym("d", nd)
        x_next = cs.repmat(cs.sum1(x) + cs.sum1(u) + cs.sum1(d), nx, 1)
        F = cs.Function("F", [x, u, d], [x_next])
        scmpc = SCMPC[cs.SX](Nlp(sym_type=sym_tpe), K, N, N // 2, shooting="multi")
        scmpc.state("x1", nx // 2)
        scmpc.state("x2", nx - (nx // 2))
        scmpc.action("u1", nu // 3)
        scmpc.action("u2", nu - (nu // 3))
        scmpc.disturbance("d", nd)
        scmpc.set_nonlinear_dynamics(F)

        self.assertIn("dyn", scmpc.constraints.keys())
        self.assertEqual(scmpc.nlp.ng, (1 + N) * nx * K)

    @parameterized.expand([("SX",), ("MX",)])
    def test_dynamics__in_singleshooting__creates_state_trajectories(self, sym_type):
        nx, nu, nd, N, K = np.random.randint(4, 20, size=5)
        x, u, d = cs.SX.sym("x", nx), cs.SX.sym("u", nu), cs.SX.sym("d", nd)
        x_next = cs.repmat(cs.sum1(x) + cs.sum1(u) + cs.sum1(d), nx, 1)
        F = cs.Function("F", [x, u, d], [x_next])
        scmpc = SCMPC[cs.SX](Nlp(sym_type=sym_type), K, N, N // 2, shooting="single")
        scmpc.state("x1", nx // 2)
        scmpc.state("x2", nx - (nx // 2))
        scmpc.action("u1", nu // 3)
        scmpc.action("u2", nu - (nu // 3))
        scmpc.disturbance("d", nd)
        scmpc.set_nonlinear_dynamics(F)

        self.assertNotIn("dyn", scmpc.constraints.keys())
        self.assertEqual(scmpc.nlp.ng, 0)
        for i in range(K):
            self.assertIn(_n("x1", i), scmpc.states.keys())
            self.assertIn(_n("x2", i), scmpc.states.keys())
            self.assertEqual(scmpc.states[_n("x1", i)].shape, (nx // 2, N + 1))
            self.assertEqual(scmpc.states[_n("x2", i)].shape, (nx - (nx // 2), N + 1))

    @parameterized.expand(product((True, False), ("multi", "single")))
    def test_constraint_from_single__creates_constraint_for_all_scenarios(
        self, soft: bool, shooting: str
    ):
        N, K = 10, np.random.randint(2, 20)
        scmpc = SCMPC[cs.SX](Nlp(sym_type="SX"), K, N, shooting=shooting)
        x, _, _ = scmpc.state("x", 2)
        u, _ = scmpc.action("u", 2)
        d, _ = scmpc.disturbance("d", 4)
        scmpc.set_nonlinear_dynamics(lambda x, u, d: x + u + d[:2] - d[2:])

        lhs = cs.vertcat(cs.sum2(cs.exp(x)), cs.sum2(cs.log(u)))
        rhs = cs.sum2(cs.sin(d))
        expr = lhs - rhs
        out = scmpc.constraint_from_single("c", lhs, ">=", rhs, soft=soft)

        self.assertEqual(scmpc.nh, expr.shape[0] * K)
        exprs, lams = out[:2]
        self.assertTrue(len(exprs) == K)
        self.assertTrue(len(lams) == K)
        self.assertTrue(all(o.shape == expr.shape for o in exprs))
        self.assertTrue(all(o.shape == expr.shape for o in lams))
        if soft:
            slack, slacks = out[2:]
            self.assertTrue(len(slacks) == K)
            self.assertTrue(all(o.shape == expr.shape for o in slacks))
            self.assertTrue(slack.shape == expr.shape)

        for i, expr_i in enumerate(out[0]):
            x_i = scmpc.states_i(i)["x"]
            d_i = scmpc.disturbances_i(i)["d"]
            vars_i = [x_i, d_i, u]
            if soft:
                s_i = scmpc.slacks_i(i)["slack_c"]
                vars_i.append(s_i)

            for var in cs.symvar(expr_i):
                found = False
                for arr in vars_i:
                    for j, k in product(range(arr.shape[0]), range(arr.shape[1])):
                        if cs.is_equal(var, arr[j, k]):
                            found = True
                            break
                self.assertTrue(found, f"Variable {var} not found in scenario {i}.")

    def test_minimize_from_single__creates_objective_for_all_scenarios(self):
        N, K = 10, np.random.randint(2, 20)
        scmpc = SCMPC[cs.SX](Nlp(sym_type="SX"), K, N, shooting="multi")
        x, _, _ = scmpc.state("x", 2)
        u, _ = scmpc.action("u", 2)
        d, _ = scmpc.disturbance("d", 4)
        lhs = cs.vertcat(cs.sum2(cs.exp(x)), cs.sum2(cs.log(u)))
        rhs = cs.sum2(cs.sin(d))
        _, _, s, _ = scmpc.constraint_from_single("c", lhs, ">=", rhs, soft=True)

        def objective(x, u, d, s):
            return (
                cs.sumsqr(x)
                + cs.sum1(cs.sum2(cs.log(cs.fabs(u))))
                + cs.sum1(cs.sum2(cs.sqrt(cs.exp(d))))
                + cs.sum1(cs.sum2(s))
            )

        scmpc.minimize_from_single(objective(x, u, d, s))

        # symbolical check
        for var in cs.symvar(scmpc.nlp.f):
            found = any(
                cs.is_equal(var, arr[j, k])
                for i in range(K)
                for arr in (
                    scmpc.states_i(i)["x"],
                    u,
                    scmpc.disturbances_i(i)["d"],
                    scmpc.slacks_i(i)["slack_c"],
                )
                for j, k in product(range(arr.shape[0]), range(arr.shape[1]))
            )
            self.assertTrue(found, f"Variable {var} not found.")

        # numerical check
        numerical_J = 0.0
        u_ = np.random.randn(*u.shape)
        symbolical_J = subsevalf(scmpc.nlp.f, u, u_, eval=False)
        for i in range(K):
            x_i = scmpc.states_i(i)["x"]
            d_i = scmpc.disturbances_i(i)["d"]
            s_i = scmpc.slacks_i(i)["slack_c"]
            x_i_ = np.random.randn(*x_i.shape)
            d_i_ = np.random.randn(*d_i.shape)
            s_i_ = np.random.randn(*s_i.shape)
            numerical_J += objective(x_i_, u_, d_i_, s_i_) / K
            symbolical_J = subsevalf(
                symbolical_J, [x_i, d_i, s_i], [x_i_, d_i_, s_i_], eval=i == K - 1
            )
        np.testing.assert_almost_equal(symbolical_J, numerical_J)


if __name__ == "__main__":
    unittest.main()
